# -*- coding: utf-8 -*-
"""
Created on Mon Mar 22 14:31:51 2021

@author: WUW9
"""

import os
import re
import glob
import numpy as np
import scipy.io as sio
import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import DataLoader
from torch.autograd import Variable
from torch.utils.data import Dataset
import ctlib
import cv2
from torch.autograd import Function
systemMat = torch.sparse.FloatTensor()
cuda = True if torch.cuda.is_available() else False

class fan_ed(Function):
    def __init__(self, views, dets, width, height, dImg, dDet, dAng, s2r, d2r, binshift):
        self.options = torch.Tensor([views, dets, width, height, dImg, dDet, dAng, s2r, d2r, binshift])
        self.options = self.options.double().cuda()

    def forward(self, image):
        return ctlib.projection(image, self.options, 1)
    
    def backward(self, projection):
        return ctlib.fbp(projection, self.options, 1)
    

class trainset_loader(Dataset):
    def __init__(self):
        self.files_A = sorted(glob.glob('./train/label/' + '*.mat'))
        #print(0)
        
    def __getitem__(self, index):
        file_A = self.files_A[index]
        input_data = sio.loadmat(file_A)['sinofull']
        input_data = cv2.GaussianBlur(input_data,(3,3),0)
        input_data = cv2.GaussianBlur(input_data,(3,3),0)
        file_B = file_A.replace('label','projection')
        input_data = torch.FloatTensor(input_data.reshape(1,880,2200))
        input_data=input_data.permute(0,2,1)
        return input_data, file_B

    def __len__(self):
        return len(self.files_A)

class testset_loader(Dataset):
    def __init__(self):
        self.files_A = sorted(glob.glob('./test/label/' + '*.mat'))
        
    def __getitem__(self, index):
        file_A = self.files_A[index]
        file_B = file_A.replace('label','projection')
        input_data = sio.loadmat(file_A)['sinofull']
        input_data = cv2.GaussianBlur(input_data,(3,3),0)
        input_data = cv2.GaussianBlur(input_data,(3,3),0)
        input_data = torch.FloatTensor(input_data.reshape(1,880,2200))
        input_data=input_data.permute(0,2,1)
        return input_data, file_B

    def __len__(self):
        return len(self.files_A)


if __name__ == "__main__":
    
    # loader=trainset_loader()
    imagesize=256
    pixelsize=0.009*2
    batch_size=4
    train_data = DataLoader(trainset_loader(), batch_size, shuffle=1, num_workers=0)
    test_data = DataLoader(testset_loader(), batch_size, shuffle=1, num_workers=0)
    projector = fan_ed(2304, 736, imagesize, imagesize, pixelsize, 0.0012, 2*3.14159/2304, 5.95, 4.906, -0.0013)
    projector2 = fan_ed(2304, 736, 512, 512, 0.009, 0.0012, 2*3.14159/2304, 5.95, 4.906, -0.0013)
    projector1 = fan_ed(2200, 880, imagesize, imagesize, pixelsize, 0.0011, 2*3.14159/2200, 5.3852, 0, -0.0013)
    projector3 = fan_ed(2200, 880, 512, 512, 0.009, 0.0011, 2*3.14159/2200, 5.3852, 0, -0.0013)
    projector_sparse = fan_ed(48, 880, imagesize, imagesize,pixelsize,0.0011, 2*3.14159/2200*30, 5.3852, 0, -0.0013)
    projector_sparseinter = fan_ed(144, 880, imagesize, imagesize,pixelsize,0.0011, 2*3.14159/2200*10, 5.3852, 0, -0.0013)
    a=torch.ones((batch_size,1,imagesize,imagesize))
    a=a.double()
    proj_one = projector_sparse.forward(a.cuda())
    imageone = projector_sparse.backward(proj_one)
    ata=torch.sum((imageone)*(imageone))/torch.sum((a)*(a))
    for batch_index, data in enumerate(train_data):
        input_data, file_name = data
        input_data[input_data<0]=0
        input_data=input_data
        input_data = input_data.double()
        if cuda:
            input_data = input_data.cuda()
            # proj_sparse = proj_sparse.cuda()
        imagef = projector1.backward(input_data)
        imagef2 = projector3.backward(input_data)
        projfull = projector1.forward(imagef)
        projfull1=projfull.cpu().numpy()
        proj_sparse= projfull1[:,:,0:-760:30,:]
        proj_sparseinter= projfull1[:,:,0:-760:10,:]
        proj_sparse=torch.FloatTensor(proj_sparse)
        proj_sparse = proj_sparse.double()
        proj_sparseinter=torch.FloatTensor(proj_sparseinter)
        proj_sparseinter = proj_sparseinter.double()
        if cuda:
            proj_sparse = proj_sparse.cuda()
            proj_sparseinter = proj_sparseinter.cuda()
        imagesparse = projector_sparse.backward(proj_sparse)
        for i in range(proj_sparse.size(0)):
            temp = proj_sparse[i].cpu().numpy().reshape(48,880)
            temp0= proj_sparseinter[i].cpu().numpy().reshape(144,880)
            temp1 = imagef[i].cpu().numpy().reshape(imagesize,imagesize)
            temp2 = imagesparse[i].cpu().numpy().reshape(imagesize,imagesize)
            temp3 = imagef2[i].cpu().numpy().reshape(512,512)
            temp1[temp1<0]=0
            temp2[temp2<0]=0
            temp3[temp3<0]=0
            sio.savemat(file_name[i], {'data':temp,'data1':temp0,'imagef':temp1,'imagesparse':temp2,'imagef2':temp3})
    for batch_index, data in enumerate(test_data):
        input_data, file_name = data
        input_data[input_data<0]=0
        
        input_data = input_data.double()
        if cuda:
            input_data = input_data.cuda()
        imagef = projector1.backward(input_data)
        imagef2 = projector3.backward(input_data)
        projfull = projector1.forward(imagef)
        
        projfull1=projfull.cpu().numpy()
        proj_sparse= projfull1[:,:,0:-760:30,:]        #48*30+760=2200
        proj_sparseinter= projfull1[:,:,0:-760:10,:]    
        proj_sparse=torch.FloatTensor(proj_sparse)
        proj_sparse = proj_sparse.double()
        proj_sparseinter=torch.FloatTensor(proj_sparseinter)
        proj_sparseinter = proj_sparseinter.double()
        if cuda:
            proj_sparse = proj_sparse.cuda()
            proj_sparseinter = proj_sparseinter.cuda()
        imagesparse =projector_sparse.backward(proj_sparse)
        for i in range(proj_sparse.size(0)):
            temp = proj_sparse[i].cpu().numpy().reshape(48,880)
            temp0= proj_sparseinter[i].cpu().numpy().reshape(144,880)
            temp1 = imagef[i].cpu().numpy().reshape(imagesize,imagesize)
            temp2 = imagesparse[i].cpu().numpy().reshape(imagesize,imagesize)
            temp3 = imagef2[i].cpu().numpy().reshape(512,512)
            temp1[temp1<0]=0
            temp2[temp2<0]=0
            sio.savemat(file_name[i], {'data':temp,'data1':temp0,'imagef':temp1,'imagesparse':temp2,'imagef2':temp3})
