# --------------------------------------------------------
# MISTNet
# School of BME, SYSU
# Written by panjy (panjy59@mail2.sysu.edu.cn) 
# --------------------------------------------------------
import torch
import torch.nn as nn
from torch.autograd import Function
import torchvision.models as models
import ctlib

from os.path import join as pjoin
from scipy import ndimage
import torch.nn.functional as F

import SwinRec
from SwinRec import SwinRecNet
#--------------------------------------------------------------------------------------------------------



class prj_module(nn.Module):
    def __init__(self, options):
        super(prj_module, self).__init__()  
        self.options = nn.Parameter(options, requires_grad=False)
        self.weight = nn.Parameter(torch.FloatTensor(1), requires_grad=True)
        self.weight.data.fill_(0.05)
    def forward(self, input_data, proj):
        
        return prj_fun.apply(input_data, self.weight, proj, self.options)

class prj_fun(Function):
    @staticmethod
    def forward(self, input_data, weight, proj, options):
        
        temp = ctlib.projection(input_data, options,1) - proj
        intervening_res = ctlib.fbp(temp, options,1)
        self.save_for_backward(intervening_res, weight, options)
        out = input_data + weight * intervening_res
        return out

    @staticmethod
    def backward(self, grad_output):
        intervening_res, weight, options = self.saved_tensors
        temp = ctlib.projection(grad_output, options,1)
        temp = ctlib.fbp(temp, options,1)
        grad_input = grad_output - weight * temp
        temp = intervening_res * grad_output
        grad_weight = - temp.sum().view(-1)
        return grad_input, grad_weight, None, None
    
class prj_module0(nn.Module):
    def __init__(self, options):
        super(prj_module0, self).__init__()  
        self.weight = nn.Parameter(torch.Tensor(1))
        self.options = nn.Parameter(options, requires_grad=False)
        
    def forward(self,  proj):        
        return prj_fun0.apply(  proj, self.options)

class prj_fun0(Function):
    @staticmethod
    def forward(self, proj, options):
        intervening_res = ctlib.fbp(proj, options,1)
        self.save_for_backward(intervening_res, options)
        out =  intervening_res
        return out

    @staticmethod
    def backward(self, grad_output):
        intervening_res, options = self.saved_tensors
        temp = ctlib.projection(grad_output, options,1)
        
        grad_input =  temp
        return grad_input, None, None


class prj_module00(nn.Module):
    def __init__(self, options):
        super(prj_module00, self).__init__()  
        self.options = nn.Parameter(options, requires_grad=False)
        
    def forward(self, input_data):        
        return prj_fun00.apply(input_data, self.options)

class prj_fun00(Function):
    @staticmethod
    def forward(self, input_data, options):
        temp=ctlib.projection(input_data, options,1)
        intervening_res = temp
        self.save_for_backward(intervening_res, options)
        out =  intervening_res
        return out

    @staticmethod
    def backward(self, grad_output):
        intervening_res, options = self.saved_tensors
        temp = ctlib.fbp(grad_output, options,1)
       
        grad_input =  temp


        return grad_input,None, None
    
    
class prj_module1(nn.Module):
    def __init__(self, options):
        super(prj_module1, self).__init__()  
        self.options = nn.Parameter(options, requires_grad=False)
        
    def forward(self, input_data, proj, projinter):        
        return prj_fun1.apply(input_data,  proj, projinter, self.options)

class prj_fun1(Function):
    @staticmethod
    def forward(self, input_data, proj, projground, options):
        temp1=ctlib.projection(input_data, options,1)
        temp1[:,:,0:-1:3,:]=projground
        temp = ( proj-temp1)# ata=1.5624
        intervening_res = ctlib.fbp(temp, options,1)
        self.save_for_backward(intervening_res, options)
        out =  intervening_res
        return out, temp1

    @staticmethod
    def backward(self, grad_output,grad_temp1):
        intervening_res, options = self.saved_tensors
        temp = ctlib.projection(grad_output, options,1)
        grad_temp1 = ctlib.projection(grad_temp1, options,1)
        temp = ctlib.fbp(temp, options,1)
        grad_input =  -temp


        return grad_input,grad_temp1, None, None


class conv_block(nn.Module):
    def __init__(self,ch_in,ch_out):
        super(conv_block,self).__init__()
        self.conv = nn.Sequential(
            nn.Conv2d(ch_in, ch_out, kernel_size=3,stride=1,padding=1,bias=True),
            nn.BatchNorm2d(ch_out),
            nn.LeakyReLU(inplace=True),
            nn.Conv2d(ch_out, ch_out, kernel_size=3,stride=1,padding=1,bias=True),
            nn.BatchNorm2d(ch_out),
            nn.LeakyReLU(inplace=True)
        )

    def forward(self,x):
        x = self.conv(x)
        return x

class up_conv(nn.Module):
    def __init__(self,ch_in,ch_out):
        super(up_conv,self).__init__()
        self.up = nn.Sequential(
            nn.Upsample(scale_factor=2),
            nn.Conv2d(ch_in,ch_out,kernel_size=3,stride=1,padding=1,bias=True),
		    nn.BatchNorm2d(ch_out),
			nn.LeakyReLU(inplace=True)
        )

    def forward(self,x):
        x = self.up(x)
        return x


#---------------------------------------------------------------------------------------------------------------------------
class SobelConv2d(nn.Module):

    def __init__(self, in_channels, out_channels, kernel_size=3, stride=1,
                 padding=0, dilation=1, groups=1, bias=True, requires_grad=True):
        assert kernel_size % 2 == 1, 'SobelConv2d\'s kernel_size must be odd.'
        assert out_channels % 4 == 0, 'SobelConv2d\'s out_channels must be a multiple of 4.'
        assert out_channels % groups == 0, 'SobelConv2d\'s out_channels must be a multiple of groups.'

        super(SobelConv2d, self).__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding
        self.dilation = dilation
        self.groups = groups

        # In non-trainable case, it turns into normal Sobel operator with fixed weight and no bias.
        self.bias = bias if requires_grad else False

        if self.bias:
            self.bias = nn.Parameter(torch.zeros(size=(out_channels,), dtype=torch.float32), requires_grad=True)
        else:
            self.bias = None

        self.sobel_weight = nn.Parameter(torch.zeros(
            size=(out_channels, int(in_channels / groups), kernel_size, kernel_size)), requires_grad=False)

        # Initialize the Sobel kernal
        kernel_mid = kernel_size // 2
        for idx in range(out_channels):
            if idx % 4 == 0:
                self.sobel_weight[idx, :, 0, :] = -1
                self.sobel_weight[idx, :, 0, kernel_mid] = -2
                self.sobel_weight[idx, :, -1, :] = 1
                self.sobel_weight[idx, :, -1, kernel_mid] = 2
            elif idx % 4 == 1:
                self.sobel_weight[idx, :, :, 0] = -1
                self.sobel_weight[idx, :, kernel_mid, 0] = -2
                self.sobel_weight[idx, :, :, -1] = 1
                self.sobel_weight[idx, :, kernel_mid, -1] = 2
            elif idx % 4 == 2:
                self.sobel_weight[idx, :, 0, 0] = -2
                for i in range(0, kernel_mid + 1):
                    self.sobel_weight[idx, :, kernel_mid - i, i] = -1
                    self.sobel_weight[idx, :, kernel_size - 1 - i, kernel_mid + i] = 1
                self.sobel_weight[idx, :, -1, -1] = 2
            else:
                self.sobel_weight[idx, :, -1, 0] = -2
                for i in range(0, kernel_mid + 1):
                    self.sobel_weight[idx, :, kernel_mid + i, i] = -1
                    self.sobel_weight[idx, :, i, kernel_mid + i] = 1
                self.sobel_weight[idx, :, 0, -1] = 2
        
        # Define the trainable sobel factor
        if requires_grad:
            self.sobel_factor = nn.Parameter(torch.ones(size=(out_channels, 1, 1, 1), dtype=torch.float32),
                                             requires_grad=True)
        else:
            self.sobel_factor = nn.Parameter(torch.ones(size=(out_channels, 1, 1, 1), dtype=torch.float32),
                                             requires_grad=False)

    def forward(self, x):
        if torch.cuda.is_available():
            self.sobel_factor = self.sobel_factor.cuda()
            if isinstance(self.bias, nn.Parameter):
                self.bias = self.bias.cuda()

        sobel_weight = self.sobel_weight * self.sobel_factor

        if torch.cuda.is_available():
            sobel_weight = sobel_weight.cuda()

        out = F.conv2d(x, sobel_weight, self.bias, self.stride, self.padding, self.dilation, self.groups)

        return out


class EdgeEnhanceREC(nn.Module):

    def __init__(self, in_ch=1, out_ch=32, sobel_ch=32):
        super(EdgeEnhanceREC, self).__init__()

        self.conv_sobel = SobelConv2d(in_ch, sobel_ch, kernel_size=3, stride=1, padding=1, bias=True)

        self.conv_p1 = nn.Conv2d(in_ch + sobel_ch, out_ch, kernel_size=1, stride=1, padding=0)
        self.conv_f1 = nn.Conv2d(out_ch, out_ch, kernel_size=3, stride=1, padding=1)

        self.conv_p2 = nn.Conv2d(in_ch + sobel_ch + out_ch, out_ch, kernel_size=1, stride=1, padding=0)
        self.conv_f2 = nn.Conv2d(out_ch, out_ch, kernel_size=3, stride=1, padding=1)

        self.conv_p3 = nn.Conv2d(in_ch + sobel_ch + out_ch, out_ch, kernel_size=1, stride=1, padding=0)
        self.conv_f3 = nn.Conv2d(out_ch, out_ch, kernel_size=3, stride=1, padding=1)

        self.conv_p4 = nn.Conv2d(in_ch + sobel_ch + out_ch, out_ch, kernel_size=1, stride=1, padding=0)
        self.conv_f4 = nn.Conv2d(out_ch, out_ch, kernel_size=3, stride=1, padding=1)

        self.conv_p5 = nn.Conv2d(in_ch + sobel_ch + out_ch, out_ch, kernel_size=1, stride=1, padding=0)
        self.conv_f5 = nn.Conv2d(out_ch, out_ch, kernel_size=3, stride=1, padding=1)

        self.conv_p6 = nn.Conv2d(in_ch + sobel_ch + out_ch, out_ch, kernel_size=1, stride=1, padding=0)
        self.conv_f6 = nn.Conv2d(out_ch, out_ch, kernel_size=3, stride=1, padding=1)

        self.conv_p7 = nn.Conv2d(in_ch + sobel_ch + out_ch, out_ch, kernel_size=1, stride=1, padding=0)
        self.conv_f7 = nn.Conv2d(out_ch, out_ch, kernel_size=3, stride=1, padding=1)

        self.conv_p8 = nn.Conv2d(in_ch + sobel_ch + out_ch, out_ch, kernel_size=1, stride=1, padding=0)
        self.conv_f8 = nn.Conv2d(out_ch, in_ch, kernel_size=3, stride=1, padding=1)

        self.relu = nn.LeakyReLU()

    def forward(self, x):
        out_0 = self.conv_sobel(x)
        out_0 = torch.cat((x, out_0), dim=-3)

        out_1 = self.relu(self.conv_p1(out_0))
        out_1 = self.relu(self.conv_f1(out_1))
        out_1 = torch.cat((out_0, out_1), dim=-3)

        out_2 = self.relu(self.conv_p2(out_1))
        out_2 = self.relu(self.conv_f2(out_2))
        out_2 = torch.cat((out_0, out_2), dim=-3)

        out_3 = self.relu(self.conv_p3(out_2))
        out_3 = self.relu(self.conv_f3(out_3))
        out_3 = torch.cat((out_0, out_3), dim=-3)

        out_4 = self.relu(self.conv_p4(out_3))
        out_4 = self.relu(self.conv_f4(out_4))
        out_4 = torch.cat((out_0, out_4), dim=-3)

        out_5 = self.relu(self.conv_p5(out_4))
        out_5 = self.relu(self.conv_f5(out_5))
        out_5 = torch.cat((out_0, out_5), dim=-3)

        out_6 = self.relu(self.conv_p6(out_5))
        out_6 = self.relu(self.conv_f6(out_6))
        out_6 = torch.cat((out_0, out_6), dim=-3)

        out_7 = self.relu(self.conv_p7(out_6))
        out_7 = self.relu(self.conv_f7(out_7))
        out_7 = torch.cat((out_0, out_7), dim=-3)

        out_8 = self.relu(self.conv_p8(out_7))
        out_8 = self.conv_f8(out_8)

        out = self.relu(x + out_8)

        return out



#-----------------------------
class IterBlocku(nn.Module):                          
    def __init__(self,img_ch=1,output_ch=1):
        super(IterBlocku,self).__init__()
        numberchannel=32;
        self.Maxpool = nn.MaxPool2d(kernel_size=2,stride=2)

        self.Conv1 = conv_block(ch_in=1,ch_out=numberchannel)
        self.Conv2 = conv_block(ch_in=numberchannel,ch_out=2*numberchannel)
        self.Conv3 = conv_block(ch_in=2*numberchannel,ch_out=4*numberchannel)
        self.Conv4 = conv_block(ch_in=4*numberchannel,ch_out=8*numberchannel)
        self.Conv5 = conv_block(ch_in=8*numberchannel,ch_out=16*numberchannel)
        
        self.Up5 = up_conv(ch_in=16*numberchannel,ch_out=8*numberchannel)
        self.Up_conv5 = conv_block(ch_in=16*numberchannel, ch_out=8*numberchannel)
     
        self.Up4 = up_conv(ch_in=8*numberchannel,ch_out=4*numberchannel)
        self.Up_conv4 = conv_block(ch_in=8*numberchannel, ch_out=4*numberchannel)
        
        self.Up3 = up_conv(ch_in=4*numberchannel,ch_out=2*numberchannel)
        self.Up_conv3 = conv_block(ch_in=4*numberchannel, ch_out=2*numberchannel)
        
        self.Up2 = up_conv(ch_in=2*numberchannel,ch_out=numberchannel)
        self.Up_conv2 = conv_block(ch_in=2*numberchannel, ch_out=numberchannel)

        self.Conv_1x1 = nn.Conv2d(numberchannel,1,kernel_size=1,stride=1,padding=0)
        


    def forward(self,x):
        # encoding path
        
        x1 = self.Conv1(x)
         
        x2 = self.Maxpool(x1)
        x2 = self.Conv2(x2)
        
        x3 = self.Maxpool(x2)
        x3 = self.Conv3(x3)
        
        
        x4 = self.Maxpool(x3)
        x4 = self.Conv4(x4)
        
        
        x5 = self.Maxpool(x4)
        x5 = self.Conv5(x5)
       
        d5 = self.Up5(x5)
        d5 = torch.cat((x4,d5),dim=1)
        d5 = self.Up_conv5(d5)
        
        d4 = self.Up4(d5)
        d4 = torch.cat((x3,d4),dim=1)
        d4 = self.Up_conv4(d4)

        d3 = self.Up3(d4)
        d3 = torch.cat((x2,d3),dim=1)
        d3 = self.Up_conv3(d3)

        d2 = self.Up2(d3)
        d2 = torch.cat((x1,d2),dim=1)
        d2 = self.Up_conv2(d2)

        d1 = self.Conv_1x1(d2)
        output=d1+x
        
        return output




class IterBlock(nn.Module):
    def __init__(self, options, options1):
        super(IterBlock, self).__init__()
        self.lambda1 = torch.nn.Parameter(torch.FloatTensor(1), requires_grad=True)
        self.miu = torch.nn.Parameter(torch.FloatTensor(1), requires_grad=True)
        self.lambda1.data.fill_(0.1)
        self.miu.data.fill_(0.0)
        self.block = prj_module(options)
        self.block0 = prj_module0(options)
        self.block1 = prj_module0(options1)
        self.block2 = IterBlocku()                        
        self.block80= SwinRecNet()
        self.block90= EdgeEnhanceREC()
        self.block5 = prj_module00(options1)
        
        
    
    def forward(self, input_data, proj, proj1):
        proj2 = self.block2(proj1)
        proj2[:,:,0:-1:3,:]=proj
        tmp1=self.block1(proj2)
        tmp2=self.block90(tmp1)
        proj3 = self.block5(tmp2)
        proj4 = self.block2((proj2-proj3))
        proj4[:,:,0:-1:3,:]=proj-proj3[:,:,0:-1:3,:]
        tmp4=self.block1(proj4)
        tmp4=self.block2(tmp4)
        output1=tmp4+tmp2
        output1,output2 = self.block80(output1)
        projout=self.block5(output1)
      
        return output1,projout,output2

class MISTNET(nn.Module):
    def __init__(self, block_num, **kwargs):
        super(MISTNET, self).__init__()
        views = kwargs['views']
        dets = kwargs['dets']
        width = kwargs['width']
        height = kwargs['height']
        dImg = kwargs['dImg']
        dDet = kwargs['dDet']
        dAng = kwargs['dAng']
        s2r = kwargs['s2r']
        d2r = kwargs['d2r']
        binshift = kwargs['binshift']
        scale = kwargs['scale']
        options = torch.Tensor([views, dets, width, height, dImg, dDet, dAng, s2r, d2r, binshift])
        options1 = torch.Tensor([views*scale, dets, width, height, dImg, dDet, dAng/scale, s2r, d2r, binshift])
        self.block = nn.ModuleList([IterBlock(options, options1) for i in range(int(block_num))])
        
         
    def forward(self, input_data, proj):
        x = input_data
        proj1= nn.functional.interpolate(proj,size=[144,880],mode='nearest', align_corners=None)
        for index, module in enumerate(self.block):
            x, proj1,output2 = module(x,proj,proj1)
        
            
        return x, proj1,output2
